local localfuncs = (select(2,...))

local templates = {}
localfuncs.templates = templates

localfuncs.latestrev = function()
	local rev = "beta-1"
	while localfuncs.patchnotes[rev] do
		rev = localfuncs.patchnotes[rev][2]
	end
	return rev
end

local function tcopy(intable,outtable,overwrite,written)
	--written is an internal table to track references to already copied tables and keep those intact
	--create outtable if it doesn't exist
	outtable = outtable or {}
	--create written if this is the first call (user call)
	written = written or {}
	--add copied table to written
	written[tostring(intable)] = outtable
	--loop through data
	for key, data in pairs(intable) do
		--if overwrite or not exists, write value
		if (not outtable[key]) or overwrite then
			if type(data) == "table" then
				outtable[key] = written[tostring(data)] or tcopy(data)
			else
				outtable[key] = data
			end
		end
	end
	return outtable
end


-- GLOBALS: FirstLootEver_Items, AlertFrame_AnimateIn, AlertFrame_FixAnchors, AchievementFrame_LoadUI, AchievementFrameCategories_Update, AchievementAlertFrame_GetAlertFrame, AchievementFrame_OnEvent, WatchFrame_OnEvent, AchievementFrameAchievements_OnEvent, AchievementFrame_ForceUpdate, AchievementFrameCategories_GetCategoryList, AlertFrame_OnEvent
-- GLOBALS: ACHIEVEMENTUI_CATEGORIES
-- GLOBALS: AchievementFrame, WatchFrame, AchievementFrameAchievements, AlertFrame
-- GLOBALS: GetCategoryInfo, GetCategoryNumAchievements, GetAchievementInfo, IsTrackedAchievement, GetAchievementNumCriteria, GetPreviousAchievement, GetNextAchievement, AddTrackedAchievement, RemoveTrackedAchievement, GetAchievementCriteriaInfo, GetAchievementCategory, GetAchievementLink, GetTrackedAchievements, GetNumTrackedAchievements
-- GLOBALS: SLASH_FIRSTLOOTEVER1
-- GLOBALS: _
local _G = _G
local unpack, pairs, ipairs, next, tinsert, tremove, print, type, tonumber, select, date, time, setmetatable, math, bit, table, tContains =
	  unpack, pairs, ipairs, next, tinsert, tremove, print, type, tonumber, select, date, time, setmetatable, math, bit, table, tContains
local GetRealmName, GetContainerNumSlots, GetContainerItemID, CreateFrame, GetPlayerInfoByGUID, UnitGUID, SetDesaturation =
	  GetRealmName, GetContainerNumSlots, GetContainerItemID, CreateFrame, GetPlayerInfoByGUID, UnitGUID, SetDesaturation
local BACKPACK_CONTAINER, NUM_BAG_SLOTS, ACHIEVEMENT_UNLOCKED, ACHIEVEMENT_TOOLTIP_COMPLETE, ACHIEVEMENT_TOOLTIP_IN_PROGRESS =
	  BACKPACK_CONTAINER, NUM_BAG_SLOTS, ACHIEVEMENT_UNLOCKED, ACHIEVEMENT_TOOLTIP_COMPLETE, ACHIEVEMENT_TOOLTIP_IN_PROGRESS
local GameTooltip =
	  GameTooltip

local dummyFunc = function() end
local gii = GetItemInfo
local iteminfo_cache = {}
local function GetItemInfo(i)
	if iteminfo_cache[i] then return unpack(iteminfo_cache[i]) end
	local retVals = {gii(i)}
	if not retVals[1] then
		-- automatically retry if the client might just not have cached the item yet
		retVals = {gii(i)}
	end
	if not retVals[1] then
		return nil
	end
	iteminfo_cache[i] = retVals
	return unpack(retVals)
end
localfuncs.gii = GetItemInfo

local giqc = GetItemQualityColor
local qualcolor_cache = {}
local function GetItemQualityColor(q)
	if qualcolor_cache[q] then return unpack(qualcolor_cache[q]) end
	local retvals = {giqc(q)}
	if not retvals[1] then return nil end
	qualcolor_cache[q] = retvals
	return unpack(retvals)
end
localfuncs.giqc=GetItemQualityColor

local achievements = {}

local addon = CreateFrame("Frame", "FirstLootEver")

local itemList = {}
local ignoredItems = {}
localfuncs.itemlist= function() return itemList end
local filterOptions = {}
local filterLabels = {
	[1] = "Slot",
	[2] = "Quality",
	[3] = "Type",
	[4] = "Expansion",
	[5] = "Weapon type",
}
function addon:RegisterItem(itemID, itemFilterMatches)
	local num, filter = next(itemFilterMatches, nil)
	local lastNum
	while (num and filter) do
		if filter == "" then
			itemFilterMatches[num] = nil
			num, filter = next(itemFilterMatches,lastNum)
		else
			filterOptions[num] = filterOptions[num] or {}
			if not tContains(filterOptions[num], filter) then
				tinsert(filterOptions[num], filter)
			end
			lastNum=num
			num,filter = next(itemFilterMatches,num)
		end
	end
	itemList[itemID] = itemFilterMatches
	GetItemInfo(itemID) -- just to fill the cache for later
end

function addon:IgnoreItem(itemID) -- this won't make the addon "disregard" the item, it will just prevent the "unknown item" popup from occuring
	ignoredItems[itemID] = true
end

function addon:RegenerateItemCategoryListings()
end

function addon:RegisterAchievement(id, name, desc, icon, checkFunc, criteria, category, prevAch, nextAch)
	self.namespace:RegisterAchievement(id, name, desc, icon, criteria, category, prevAch, nextAch)
	achievements[id] = checkFunc
end

function addon:RegisterAchievementFromTemplate(id, name, desc, icon, category, prevAch, nextAch, template, ...)
	local checkFunc, criteria = templates[template](id,...)
	self:RegisterAchievement(id, name, desc, icon, checkFunc, criteria, category, prevAch, nextAch)
end

function addon:HasAchievement(id)
	return self.namespace:HasAchievement(id)
end

function addon:AwardAchievement(id)
	self.namespace:AwardAchievement(id)
end

function addon:DisplayOrQueueAlert(id)
	local link = (select(2,GetItemInfo(id)))
	if link then
		print("First looted: "..link)
	end
	if not (itemList[id] or ignoredItems[id]) then
		StaticPopup_Show("FLE_UNKNOWNITEM",id)
	end
end

function addon:CheckAchievementEligibility(id)
	for avId, checkFunc in pairs(achievements) do
		if not self:HasAchievement(avId) then
			if checkFunc(self, id, avId) then
				self:AwardAchievement(avId)
			end
		end
	end
	AchievementFrameAchievements_OnEvent(AchievementFrameAchievements, "CRITERIA_UPDATE")
	AchievementFrameAchievements_OnEvent(AchievementFrameAchievements, "TRACKED_ACHIEVEMENT_UPDATE")
	WatchFrame_OnEvent(WatchFrame, "TRACKED_ACHIEVEMENT_UPDATE")
end

local mainFrame = CreateFrame("Frame", "FirstLootEver_Main", UIParent)
mainFrame:SetFrameStrata("DIALOG")
mainFrame:SetSize(480, 544)
mainFrame:SetBackdrop({
	bgFile="Interface\\DialogFrame\\UI-DialogBox-Background",
	edgeFile="Interface\\DialogFrame\\UI-DialogBox-Border",
	tileSize=32,
	tile=true,
	insets={
		top=12,
		bottom=12,
		right=12,
		left=12,
	},
})
mainFrame.header = CreateFrame("Button", nil, mainFrame)
mainFrame.header:SetNormalTexture("Interface\\DialogFrame\\UI-DialogBox-Header")
mainFrame.header:SetSize(239.4,61.2)
mainFrame.header:SetHitRectInsets(90.72,90.72,12.96,41.76)
mainFrame.header.text = mainFrame.header:CreateFontString()
mainFrame.header.text:SetPoint("TOP",0,-12.16)
mainFrame.header.text:SetFont("Fonts\\FRIZQT__.TTF", 14, "")
mainFrame.header.text:SetTextColor(1,0.82,0)
mainFrame.header.text:SetText("Item lists")
mainFrame.header:SetMovable(true)
mainFrame.header:SetScript("OnMouseDown", function(self)
	self:StartMoving()
end)
mainFrame.header:SetScript("OnMouseUp", function(self)
	self:StopMovingOrSizing()
end)
mainFrame.header:SetPoint("CENTER",UIParent,"CENTER",0,272)
mainFrame:SetPoint("TOP", mainFrame.header, "TOP", 0, -23.28)
--mainFrame:SetScale(1.6)
mainFrame:Hide()

mainFrame.filters = CreateFrame("Frame", nil, mainFrame)
mainFrame.filters:SetPoint("TOPLEFT", 15, -15)
mainFrame.filters:SetPoint("TOPRIGHT", -15, -15)
mainFrame.filters:SetHeight(130)

--mainFrame.currentFilter = {[1]="Chest",[2]="Epic",[3]="Cloth",[4]=false}
mainFrame.currentFilter = {}

local function matchFilter(toMatch)
	for key, value in pairs(mainFrame.currentFilter) do
		if (value == false) then
			if toMatch[key] then
				return false
			end
		else
			if not(toMatch[key] == value) then
				return false
			end
		end
	end
	return true
end

mainFrame.filters.entries = setmetatable({}, {__index=function(t,v)
	local f = CreateFrame("Frame", nil, mainFrame.filters)
	if v > 1 then
		f:SetPoint("TOPLEFT", t[v-1], "TOPRIGHT", 5, 0)
		f:SetPoint("BOTTOMLEFT", t[v-1], "BOTTOMRIGHT", 5, 0)
	else
		f:SetPoint("TOPLEFT")
		f:SetPoint("BOTTOMLEFT")
	end
	f.label = f:CreateFontString()
	f.label:SetFont("Fonts\\FRIZQT__.TTF", 12, "OUTLINE")
	f.label:SetTextColor(1,0.82,0)
	f.label:SetPoint("TOP", 0, -2)
	f.label:SetHeight(20)
	f.filters = CreateFrame("ScrollFrame", nil, f)
	f.filters:SetPoint("TOPLEFT", 2, -24)
	f.filters:SetPoint("BOTTOMRIGHT", -2, 2)
	f.filters.child = CreateFrame("Frame", nil, f.filters)
	f.filters.child:EnableMouseWheel(true)
	f.filters.child:SetScript("OnMouseWheel", function(self, delta)
		if delta == 1 then
			f.filters:SetVerticalScroll(math.max(f.filters:GetVerticalScroll()-10,0))
		else
			f.filters:SetVerticalScroll(math.min(f.filters:GetVerticalScroll()+10,f.filters:GetVerticalScrollRange()))
		end
	end)
	f.filters.child:SetWidth(f.filters:GetWidth())
	f.filters:SetScrollChild(f.filters.child)
	f.filters:SetScript("OnSizeChanged", function(self, width)
		self.child:SetWidth(width)
		self:UpdateScrollChildRect()
	end)
	f.filters.child.entries = setmetatable({},{__index=function(iT,iV)
		local iF = CreateFrame("Frame", nil, f.filters.child)
		if iV > 1 then
			iF:SetPoint("TOPLEFT", iT[iV-1], "BOTTOMLEFT", 0, -4)
			iF:SetPoint("TOPRIGHT", iT[iV-1], "BOTTOMRIGHT", 0, -4)
		else
			iF:SetPoint("TOPLEFT")
			iF:SetPoint("TOPRIGHT")
		end
		iF:SetHeight(21)
		iF.check = CreateFrame("CheckButton", nil, iF, "UICheckButtonTemplate")
		iF.check:SetSize(18,18)
		iF.check:SetPoint("LEFT", 4, 0)
		iF.check:SetScript("PostClick", function(self)
			for key, iFr in ipairs(iT) do
				if key ~= iV then
					iFr.check:SetChecked(false)
				end
			end
			self:SetChecked(true)
			mainFrame.currentFilter[f.filterLabel] = iF.optionValue
			mainFrame.list:Update()
		end)
		iF.label = iF:CreateFontString()
		iF.label:SetPoint("LEFT", iF.check, "RIGHT", 4, 0)
		iF.label:SetPoint("RIGHT", -4, 0)
		iF.label:SetWordWrap(false)
		iF.label:SetFont("Fonts\\FRIZQT__.TTF", 12, "")
		iF.label:SetTextColor(1,1,1)
		iF.label:SetJustifyH("LEFT")
		--[[iF.check:SetCheckedTexture("Interface\\Buttons\\UI-CheckBox-Check")
		iF.check:SetNormalTexture("Interface\\Buttons\\--]]
		iT[iV]=iF
		return iF
	end})
	function f:Populate(label, list, filterLabel, defaultValue)
		if not filterLabel then filterLabel = label end
		if self.filterLabel then
			mainFrame.currentFilter[filterLabel] = nil
		end
		self.filterLabel = filterLabel
		self.label:SetText(label)
		for _, btn in ipairs(self.filters.child.entries) do btn:Hide() end
		local c=1
		-- i'm making this not available for the time being. it's far too likely to just bust the client
		self.filters.child.entries[c].label:SetText("Any")
		self.filters.child.entries[c].check:SetChecked(defaultValue == nil)
		self.filters.child.entries[c]:Show()
		c=c+1
		self.filters.child.entries[c].label:SetText("None")
		self.filters.child.entries[c].optionValue = false
		self.filters.child.entries[c].check:SetChecked(defaultValue == false)
		self.filters.child.entries[c]:Show()
		c=c+1
		for _, option in ipairs(list) do
			self.filters.child.entries[c].label:SetText(option)
			self.filters.child.entries[c].optionValue = option
			self.filters.child.entries[c].check:SetChecked(defaultValue==option)
			self.filters.child.entries[c]:Show()
			c=c+1
		end
		self.filters.child:SetHeight((25*(c-1))-4)
		--self.filters:UpdateScrollChildRect()
	end
	t[v] = f
	return f
end})
function mainFrame.filters:Update()
	for _, btn in ipairs(self.entries) do
		btn:Hide()
	end
	local widthEach = (450/#filterOptions)-5
	for num, filterOpt in ipairs(filterOptions) do
		self.entries[num]:Populate(filterLabels[num], filterOpt, num, mainFrame.currentFilter[num])
		self.entries[num]:Show()
		self.entries[num]:SetWidth(widthEach)
	end
end

function mainFrame:Update()
	self.filters:Update()
	self.list:Update()
end
mainFrame:SetScript("OnShow", mainFrame.Update)

mainFrame.showAll = true

mainFrame.list = CreateFrame("ScrollFrame", nil, mainFrame)
mainFrame.list:SetPoint("BOTTOMLEFT", 15, 15)
mainFrame.list:SetPoint("BOTTOMRIGHT", -15, 15)
mainFrame.list:SetHeight(380)
mainFrame.list.child = CreateFrame("Frame", nil, mainFrame.list)
mainFrame.list.child:SetWidth(450)
mainFrame.list.child:EnableMouseWheel(true)
mainFrame.list.child:SetScript("OnMouseWheel", function(self, delta)
	if delta == 1 then
		mainFrame.list:SetVerticalScroll(math.max(mainFrame.list:GetVerticalScroll()-40,0))
	else
		mainFrame.list:SetVerticalScroll(math.min(mainFrame.list:GetVerticalScroll()+40,mainFrame.list:GetVerticalScrollRange()))
	end
end)
mainFrame.list:SetScrollChild(mainFrame.list.child)
mainFrame.list.entries = setmetatable({}, {__index=function(t,v)
	local f = CreateFrame("Frame", nil, mainFrame.list.child)
	f:SetWidth(90)
	f:SetHeight(20)
	f:EnableMouse(true)
	f:SetScript("OnEnter", function(self)
		if self.link then
			GameTooltip:ClearLines()
			GameTooltip:SetOwner(mainFrame, "ANCHOR_NONE")
			GameTooltip:ClearAllPoints()
			GameTooltip:SetPoint("TOPRIGHT", mainFrame, "TOPLEFT", -5, -155)
			GameTooltip:SetHyperlink(self.link)
			if self.obtTime then
				local obtData = date("*t", self.obtTime)
				GameTooltip:AddLine(("Item first looted on %2$d/%1$d/%3$d"):format(obtData.day, obtData.month, obtData.year),0,1,0)
			else
				GameTooltip:AddLine("Item not yet looted",0,1,0)
			end
			GameTooltip:Show()
		end
	end)
	f:SetScript("OnLeave", function()
		GameTooltip:Hide()
	end)
	f:SetBackdrop({
		bgFile="Interface\\FriendsFrame\\UI-Toast-Background",
		tileSize=12,
		edgeFile="Interface\\FriendsFrame\\UI-Toast-Border",
		tile=1,
		edgeSize=12,
		insets={
			top=5,
			bottom=5,
			right=5,
			left=5,
		},
	})
	f:SetPoint("TOPLEFT", ((v-1)%5)*90, math.floor((v-1)/5)*(-20))
	f.icon = f:CreateTexture()
	f.icon:SetPoint("LEFT", 2, 0)
	f.icon:SetSize(16,16)
	f.label = f:CreateFontString()
	f.label:SetPoint("LEFT", f.icon, "RIGHT", 5, 0)
	f.label:SetPoint("RIGHT", -2, 0)
	f.label:SetWordWrap(false)
	f.label:SetJustifyH("LEFT")
	f.label:SetFont("Fonts\\FRIZQT__.TTF", 10, "OUTLINE")
	function f:Set(itemID, hasItem)
		local name, link, quality, _, _, _, _, _, _, icon = GetItemInfo(itemID)
		if not name then
			self.link = nil
			self:Show()
			self:SetBackdropColor(0.3,0.3,0.3)
			self:SetBackdropBorderColor(0.3,0.3,0.3)
			self.label:SetTextColor(0.3,0.3,0.3)
			self.icon:SetTexture(nil)
			self.label:SetText("(#"..itemID..")")
			return
		end
		local r,g,b=GetItemQualityColor(quality)
		self.obtTime = hasItem
		self.link = link
		self.label:SetText(name)
		self.icon:SetTexture(icon)
		if hasItem then
			self:SetBackdropColor(r,g,b)
			self:SetBackdropBorderColor(r,g,b)
			self.label:SetTextColor(r,g,b)
			SetDesaturation(self.icon, false)
		else
			self:SetBackdropColor(r*0.5,g*0.5,b*0.5)
			self:SetBackdropBorderColor(r*0.5,g*0.5,b*0.5)
			self.label:SetTextColor(r*0.5,g*0.5,b*0.5)
			SetDesaturation(self.icon, true)
		end
		self:Show()
	end
	t[v] = f
	return f
end})

mainFrame.toggle = CreateFrame("Frame", nil, mainFrame)
mainFrame.toggle:SetPoint("BOTTOMRIGHT", mainFrame, "TOPRIGHT")
mainFrame.toggle:SetSize(175, 20)
mainFrame.toggle.check = CreateFrame("CheckButton", nil, mainFrame.toggle, "UICheckButtonTemplate")
mainFrame.toggle.check:SetSize(18,18)
mainFrame.toggle.check:SetPoint("LEFT", 4, 0)
mainFrame.toggle.check:SetScript("PostClick", function(self)
	mainFrame.showAll = not self:GetChecked()
	mainFrame.list:Update()
end)
mainFrame.toggle.label = mainFrame.toggle:CreateFontString()
mainFrame.toggle.label:SetPoint("LEFT", mainFrame.toggle.check, "RIGHT", 4, 0)
mainFrame.toggle.label:SetPoint("RIGHT", -4, 0)
mainFrame.toggle.label:SetWordWrap(false)
mainFrame.toggle.label:SetFont("Fonts\\FRIZQT__.TTF", 12, "")
mainFrame.toggle.label:SetTextColor(1,1,1)
mainFrame.toggle.label:SetJustifyH("LEFT")
mainFrame.toggle.label:SetText("Hide obtained")

mainFrame.toggle2 = CreateFrame("Frame", nil, mainFrame)
mainFrame.toggle2:SetPoint("BOTTOMLEFT", mainFrame, "TOPLEFT")
mainFrame.toggle2:SetSize(180,20)
mainFrame.toggle2.check = CreateFrame("CheckButton", nil, mainFrame.toggle2, "UICheckButtonTemplate")
mainFrame.toggle2.check:SetSize(18,18)
mainFrame.toggle2.check:SetPoint("LEFT", 4, 0)
mainFrame.toggle2.check:SetScript("PostClick", function(self)
	addon.savedVar.lcaSaved.announceAchievement = self:GetChecked()
end)
mainFrame.toggle2.label = mainFrame.toggle2:CreateFontString()
mainFrame.toggle2.label:SetPoint("LEFT", mainFrame.toggle2.check, "RIGHT", 4, 0)
mainFrame.toggle2.label:SetPoint("RIGHT", -4, 0)
mainFrame.toggle2.label:SetWordWrap(false)
mainFrame.toggle2.label:SetFont("Fonts\\FRIZQT__.TTF", 12, "")
mainFrame.toggle2.label:SetTextColor(1,1,1)
mainFrame.toggle2.label:SetJustifyH("LEFT")
mainFrame.toggle2.label:SetText("Announce achievements")

local MAX_ENTRY_LIMIT = 400 -- to prevent crashes
function mainFrame.list:Update()
	local entryNum=1
	local hasItems, noHasItems = {}, {}
	-- loop through all items
	for itemID, itemFilters in pairs(itemList) do
		if matchFilter(itemFilters) then
			local obtainTime = addon.savedVar.items[itemID]
			if obtainTime then
				hasItems[itemID] = obtainTime
			else
				tinsert(noHasItems, itemID)
			end
		end
	end
	if mainFrame.showAll then
		for id, obtTime in pairs(hasItems) do
			if entryNum > MAX_ENTRY_LIMIT then break end
			self.entries[entryNum]:Set(id, obtTime)
			entryNum=entryNum+1
		end
	end
	for _, id in ipairs(noHasItems) do
		if entryNum > MAX_ENTRY_LIMIT then break end
		self.entries[entryNum]:Set(id, false)
		entryNum=entryNum+1
	end
	for entCnt=entryNum, #self.entries do
		self.entries[entCnt]:Hide()
	end
	self.child:SetHeight(math.ceil((entryNum-1)/5)*20)
	self:UpdateScrollChildRect()
	self:SetVerticalScroll(0)
end

local unknownItemList = CreateFrame("Frame",nil,UIParent)
unknownItemList:SetFrameStrata("DIALOG")
--unknownItemList:Hide()
unknownItemList:SetPoint("CENTER")
unknownItemList:SetBackdrop({
	bgFile="Interface\\Tooltips\\UI-Tooltip-Background",
	tileSize=16,
	edgeFile="Interface\\Tooltips\\UI-Tooltip-Border",
	tile=1,
	edgeSize=16,
	insets={
		top=5,
		right=5,
		left=5,
		bottom=5,
	},
})
unknownItemList:SetBackdropColor(0.3,0.3,0.3)
unknownItemList:SetSize(150,400)
unknownItemList:Hide()
unknownItemList.text = unknownItemList:CreateFontString()
unknownItemList.text:SetPoint("TOPLEFT", 15, -15)
unknownItemList.text:SetPoint("TOPRIGHT", -15, -15)
unknownItemList.text:SetJustifyH("LEFT")
unknownItemList.text:SetFont("Fonts\\FRIZQT__.TTF", 12, "OUTLINE")
unknownItemList.text:SetTextColor(1,1,1)
unknownItemList.text:SetText("FLE does not know\nthese items:")
unknownItemList.hide = CreateFrame("Button", nil, unknownItemList, "UIPanelButtonTemplate")
unknownItemList.hide:SetPoint("BOTTOMLEFT", 15, 15)
unknownItemList.hide:SetPoint("BOTTOMRIGHT", -15, 15)
unknownItemList.hide:SetHeight(20)
unknownItemList.hide:SetText(HIDE)
unknownItemList.hide:SetScript("OnClick", function() unknownItemList:Hide() end)
unknownItemList.box = CreateFrame("EditBox", nil, unknownItemList)
unknownItemList.box:SetMultiLine(true)
unknownItemList.box:SetAutoFocus(false)
unknownItemList.box:SetPoint("TOPLEFT", unknownItemList.text, "BOTTOMLEFT", 0, -5)
unknownItemList.box:SetPoint("BOTTOMRIGHT", unknownItemList.hide, "TOPRIGHT", 0, 10)
unknownItemList.box:SetFont("Fonts\\FRIZQT__.TTF", 12, "OUTLINE")
unknownItemList.box:SetJustifyV("TOP")
unknownItemList.box:SetScript("OnEscapePressed", function(self) self:ClearFocus() end)
unknownItemList.box:SetScript("OnChar", function(self) self:SetText(self.origText) self:HighlightText() end)
unknownItemList.check = CreateFrame("CheckButton", nil, unknownItemList, "UICheckButtonTemplate")
unknownItemList.check:SetSize(18,18)
unknownItemList.check:SetPoint("BOTTOMLEFT", unknownItemList, "TOPLEFT")
unknownItemList.check:SetScript("PostClick", function() unknownItemList:Update() end)
unknownItemList.check.label = unknownItemList.check:CreateFontString()
unknownItemList.check.label:SetPoint("LEFT", unknownItemList.check, "RIGHT", 4, 0)
unknownItemList.check.label:SetWidth(153)
unknownItemList.check.label:SetWordWrap(false)
unknownItemList.check.label:SetFont("Fonts\\FRIZQT__.TTF", 12, "")
unknownItemList.check.label:SetTextColor(1,1,1)
unknownItemList.check.label:SetJustifyH("LEFT")
unknownItemList.check.label:SetText("Show WoWhead links")

function unknownItemList:Update()
	local oString=""
	local uIDs = {}
	local isChecked = self.check:GetChecked()
	for r, data in pairs(FirstLootEver_Items) do
		if r ~= "rev" then
			for itemID in pairs(data.items) do
				if not (uIDs[itemID] or itemList[itemID] or ignoredItems[itemID]) then
					uIDs[itemID] = true
					oString = oString..(isChecked and "http://www.wowhead.com/item=" or "")..itemID.."\n"
				end
			end
		end
	end
	self.box.origText = oString
	self.box:SetText(oString)
	self.box:HighlightText()
	self.box:SetFocus()
	self:SetWidth(isChecked and 300 or 150)
end

addon:SetScript("OnEvent", function(self, event, ...)
	if event == "ADDON_LOADED" and (...) == "FirstLootEver" then
		self:UnregisterEvent("ADDON_LOADED")
		FirstLootEver_Items = FirstLootEver_Items or {rev=localfuncs.latestrev()}
		if not FirstLootEver_Items.rev then
			FirstLootEver_Items.rev = "beta-1" -- beta-2 implemented the entry being created on savedVar initialization. if the table exists, but the entry doesn't, it's a beta-1.
		end
		-- get patch notes to display
		local displaynotes = {}
		local note_revs = {}
		local rev = FirstLootEver_Items.rev
		local oldrev = rev
		local numNote = 0
		while localfuncs.patchnotes[rev] do
			numNote = numNote+1
			for key, notes in pairs(localfuncs.patchnotes[rev][1]) do
				displaynotes[key] = displaynotes[key] or {}
				tinsert(displaynotes[key],tcopy(notes,{numNote=numNote}))
			end
			rev=localfuncs.patchnotes[rev][2]
			note_revs[numNote] = rev
		end
		FirstLootEver_Items.rev = rev
		if numNote > 0 then
			-- we have notes to display. pop up to note frame and display them.
			local isMultipleNotes = (numNote>1) -- we will display labels indicating which patch a change is from
			local noteFrame = CreateFrame("Frame", "FirstLootEver_PatchNotes", UIParent)
			noteFrame:SetSize(550,600)
			noteFrame:SetFrameStrata("DIALOG")
			noteFrame:SetBackdrop({
				bgFile="Interface\\DialogFrame\\UI-DialogBox-Background",
				edgeFile="Interface\\DialogFrame\\UI-DialogBox-Border",
				tileSize=32,
				tile=true,
				insets={
					top=12,
					bottom=12,
					right=12,
					left=12,
				},
			})
			noteFrame.header = CreateFrame("Button", nil, noteFrame)
			noteFrame.header:SetNormalTexture("Interface\\DialogFrame\\UI-DialogBox-Header")
			noteFrame.header:SetSize(239.4,61.2)
			noteFrame.header:SetHitRectInsets(90.72,90.72,12.96,41.76)
			noteFrame.header.text = noteFrame.header:CreateFontString()
			noteFrame.header.text:SetPoint("TOP",0,-12.16)
			noteFrame.header.text:SetFont("Fonts\\FRIZQT__.TTF", 14, "")
			noteFrame.header.text:SetTextColor(1,0.82,0)
			noteFrame.header.text:SetText("Item lists")
			noteFrame.header:SetMovable(true)
			noteFrame.header:SetScript("OnMouseDown", function(self)
				self:StartMoving()
			end)
			noteFrame.header:SetScript("OnMouseUp", function(self)
				self:StopMovingOrSizing()
			end)
			noteFrame.header:SetPoint("CENTER",UIParent,"CENTER",0,300)
			noteFrame:SetPoint("TOP", noteFrame.header, "TOP", 0, -23.28)
			noteFrame.closeBtn = CreateFrame("Button", nil, noteFrame, "UIPanelButtonTemplate")
			noteFrame.closeBtn:SetPoint("BOTTOMLEFT", 20, 20)
			noteFrame.closeBtn:SetPoint("BOTTOMRIGHT", -20, 20)
			noteFrame.closeBtn:SetHeight(20)
			noteFrame.closeBtn:SetText(HIDE)
			noteFrame.closeBtn:SetScript("OnClick", function() noteFrame:Hide() end)
			noteFrame.scroll = CreateFrame("ScrollFrame", nil, noteFrame)
			noteFrame.scroll:SetPoint("TOPLEFT", 10, -10)
			noteFrame.scroll:SetPoint("BOTTOMRIGHT", noteFrame.closeBtn, "TOPRIGHT", 0, 10)
			noteFrame.scroll.child = CreateFrame("Frame", nil, noteFrame.scroll)
			noteFrame.scroll.child:EnableMouseWheel(true)
			noteFrame.scroll.child:SetScript("OnMouseWheel", function(self, delta)
				if delta == 1 then
					noteFrame.scroll:SetVerticalScroll(math.max(noteFrame.scroll:GetVerticalScroll()-10,0))
				else
					noteFrame.scroll:SetVerticalScroll(math.min(noteFrame.scroll:GetVerticalScroll()+10,noteFrame.scroll:GetVerticalScrollRange()))
				end
			end)
			noteFrame.scroll.child:SetWidth(noteFrame.scroll:GetWidth())
			noteFrame.scroll:SetScrollChild(noteFrame.scroll.child)
			local totalHeight = 0
			local lastText, lastOffset
			lastText = noteFrame.scroll.child:CreateFontString()
			lastText:SetPoint("TOPLEFT", 15, -15)
			lastText:SetFont("Fonts\\FRIZQT__.TTF", 20, "OUTLINE")
			lastText:SetTextColor(0.3,0.3,1)
			lastText:SetText("FirstLootEver Patch Notes ["..oldrev.." to "..rev.."]")
			totalHeight=totalHeight+lastText:GetHeight()
			lastOffset=0
			local noteOffset = isMultipleNotes and 30 or 20
			local function formatnote(note)
				return ("- "..note):gsub("\n%s*","\n")
			end
			for _, noteKey in ipairs(localfuncs.noteorder) do
				if displaynotes[noteKey] then
					local headString = noteFrame.scroll.child:CreateFontString()
					headString:SetFont("Fonts\\FRIZQT__.TTF", 16, "OUTLINE")
					headString:SetTextColor(1,0.82,0)
					headString:SetText(noteKey)
					headString:SetPoint("TOPLEFT", lastText, "BOTTOMLEFT", (10-lastOffset), -10)
					headString:SetJustifyH("LEFT")
					totalHeight=totalHeight+10+headString:GetHeight()
					lastText,lastOffset = headString, 10
					for _, noteTable in ipairs(displaynotes[noteKey]) do
						if isMultipleNotes then
							local revString = noteFrame.scroll.child:CreateFontString()
							revString:SetFont("Fonts\\FRIZQT__.TTF", 12, "OUTLINE")
							revString:SetTextColor(1,0.82,0)
							revString:SetText(note_revs[noteTable.numNote])
							revString:SetPoint("TOPLEFT", lastText, "BOTTOMLEFT", (20-lastOffset), -7)
							revString:SetJustifyH("LEFT")
							totalHeight = totalHeight+7+revString:GetHeight()
							lastText,lastOffset=revString,20
						end
						for _, note in ipairs(noteTable) do
							local noteString = noteFrame.scroll.child:CreateFontString()
							noteString:SetIndentedWordWrap(true)
							noteString:SetFont("Fonts\\FRIZQT__.TTF", 10, "OUTLINE")
							noteString:SetTextColor(1,1,1)
							noteString:SetText(formatnote(note))
							noteString:SetPoint("TOPLEFT", lastText, "BOTTOMLEFT", (noteOffset-lastOffset), -3)
							noteString:SetWidth(noteFrame.scroll.child:GetWidth()-noteOffset-30)
							noteString:SetJustifyH("LEFT")
							totalHeight = totalHeight+3+noteString:GetHeight()
							lastText,lastOffset=noteString,noteOffset
						end
					end
				end
			end
			noteFrame.scroll.child:SetHeight(totalHeight+10)
			noteFrame.scroll:UpdateScrollChildRect()
		end
		local r = GetRealmName()
		FirstLootEver_Items[r] = FirstLootEver_Items[r] or {items={},lcaSaved={}}
		self.savedVar = FirstLootEver_Items[r]
		if not self.savedVar.items then
			-- this is a sVar from an old version, as a new one would already have the items table created
			local a = self.savedVar.achievements
			local i = {}
			for itemID, t in pairs(self.savedVar) do
				if itemID ~= "achievements" then
					i[itemID] = t
				end
			end
			table.wipe(self.savedVar)
			self.savedVar.achievements = a
			self.savedVar.items = i
			self.savedVar.trackedAchievements = {}
		end
		if not self.savedVar.achievementMemory then
			self.savedVar.achievementMemory = {}
		end
		if not self.savedVar.recentAchievementTimes then
			-- we use this for tracking FLE recent stuff
			self.savedVar.recentAchievementTimes = {}
		end
		if not self.savedVar.lcaSaved then
			-- convert old SV to LCA system
			self.savedVar.lcaSaved = {
				achievements = self.savedVar.achievements,
				recent = self.savedVar.recentAchievementTimes,
				tracked = self.savedVar.trackedAchievements,
				memory = self.savedVar.achievementMemory,
				announceAchievement = self.savedVar.announceAchievement,
				recentBlizz = {},
				wasLoadedAlready = true,
			}
			self.savedVar.achievements = nil
			self.savedVar.recentAchievementTimes = nil
			self.savedVar.trackedAchievements = nil
			self.savedVar.achievementMemory = nil
			self.savedVar.announceAchievement = nil
		end
		if self.savedVar.announceAchievement then
			self.savedVar.lcaSaved.announceAchievement = self.savedVar.announceAchievement
			self.savedVar.announceAchievement = nil
		end
		mainFrame.toggle2.check:SetChecked(self.savedVar.lcaSaved.announceAchievement)
		if FirstLootEver_BlizzAchievements then
			-- migrate into lcaSaved
			local nameRealm = (UnitName("player")).." - "..GetRealmName()
			self.savedVar.lcaSaved.recentBlizz[nameRealm] = FirstLootEver_BlizzAchievements
			self.savedVar.lcaSaved.recentBlizz[nameRealm].guid = UnitGUID("player")
			FirstLootEver_BlizzAchievements = nil
		end
		--self.achievementHistory = {blizz=FirstLootEver_BlizzAchievements,fle=self.savedVar.recentAchievementTimes}
		self.isLoaded = true
		--[[if self.scheduleBlizzAchievement then
			local curTime = time()
			for c, av in ipairs(self.scheduleBlizzAchievement) do
				tinsert(self.achievementHistory.blizz,1,{id=av, t=curTime})
				tremove(self.achievementHistory.blizz,6)
			end
			self.scheduleBlizzAchievement = nil
		end--]]
		addon.namespace = LibStub("LibCustomAchievement-1.0"):RegisterNamespace("FirstLootEver",self.savedVar.lcaSaved)
		localfuncs.registerAchievements()
		dummyFunc()
	end
	if event == "UPDATE_INSTANCE_INFO" then
		-- i randomly picked an event that is guaranteed to fire AFTER the initial BAG_UPDATE spam - just to make sure achievements aren't swallowed by the loading screen
		self:UnregisterEvent("UPDATE_INSTANCE_INFO")
		for slot=1, 19 do
			local id = GetInventoryItemID("player", slot)
			if id then -- not an empty slot
				if not self.savedVar.items[id] then
					self.savedVar.items[id] = time()
					addon:DisplayOrQueueAlert(id)
					addon:CheckAchievementEligibility(id)
				end
			end
		end
		self:RegisterEvent("BAG_UPDATE")
		self:RegisterEvent("BANKFRAME_OPENED")
		self:GetScript("OnEvent")(self,"BAG_UPDATE")
	end
	if event == "BAG_UPDATE" then
		for bag=BACKPACK_CONTAINER, NUM_BAG_SLOTS do
			for slot=1, GetContainerNumSlots(bag) do
				local id = GetContainerItemID(bag,slot)
				if id then -- not an empty slot
					if not self.savedVar.items[id] then
						self.savedVar.items[id] = time()
						addon:DisplayOrQueueAlert(id)
						addon:CheckAchievementEligibility(id)
					end
				end
			end
		end
	end
	if event == "BANKFRAME_OPENED" then
		local bag = BANK_CONTAINER
		while true do -- do it like this because first bank bag (main storage) is -1 and the others are 5++
			for slot=1, GetContainerNumSlots(bag) do
				local id = GetContainerItemID(bag,slot)
				if id then -- not an empty slot
					if not self.savedVar.items[id] then
						self.savedVar.items[id] = time()
						addon:DisplayOrQueueAlert(id)
						addon:CheckAchievementEligibility(id)
					end
				end
			end
			bag=math.max(bag+1,(NUM_BAG_SLOTS+1))
			if bag > (NUM_BAG_SLOTS+NUM_BANKBAGSLOTS) then
				break
			end
		end
	end
	--[[if event == "ACHIEVEMENT_EARNED" then
		if (select(12,GetAchievementInfo((...)))) then return end
		if not addon.isLoaded then
			addon.scheduleBlizzAchievement = addon.scheduleBlizzAchievement or {}
			tinsert(addon.scheduleBlizzAchievement,(...))
			return
		end
		tinsert(addon.achievementHistory.blizz,1,{id=(...),t=time()})
		tremove(addon.achievementHistory.blizz,6)
	end--]]
end)
addon:RegisterEvent("ADDON_LOADED")
addon:RegisterEvent("UPDATE_INSTANCE_INFO")

StaticPopupDialogs["FLE_UNKNOWNITEM"] = {
	text="You have looted an item that FLE does not know! We would appreciate if you could report this to us so we can add it to our item lists!",
	button1="Details",
	button2="Hide",
	timeout=0,
	whileDead=true,
	OnAccept=function()
		unknownItemList:Show()
		unknownItemList:Update()
	end,
}

SlashCmdList["FIRSTLOOTEVER"] = function() mainFrame[mainFrame:IsShown() and "Hide" or "Show"](mainFrame) end
SLASH_FIRSTLOOTEVER1 = "/fle"

--[===[@debug@
SlashCmdList["CHECKUNKNOWN_FLE"] = function() for iid in pairs(addon.savedVar.items) do if not (itemList[iid] or ignoredItems[iid]) then print(iid) end end end
SLASH_CHECKUNKNOWN_FLE1 = "/chufle"
--@end-debug@]===]

--[===[@alpha@
dummyFunc = function() if not addon:HasAchievement(1339) then addon:AwardAchievement(1339) end end
--@end-alpha@]===]

