assert(LibStub,"LibStub not found!")
local lib = LibStub:NewLibrary("LibCustomAchievement-1.0",4)
if not lib then return --[[does not need upgrading--]] end

-- GLOBALS: ACHIEVEMENTUI_CATEGORIES
-- GLOBALS: AchievementFrame, AchievementFrameAchievements, AchievementFrame_ForceUpdate, AchievementFrame_IsComparison, AchievementFrameAchievements_OnEvent, AchievementFrameCategories_Update, AchievementFrameCategories_GetCategoryList, AchievementFrameComparison_ForceUpdate
local isLoaded, guidLoaded, recentLoaded
local _G = _G
local bit, math, table, pairs, ipairs, tinsert, tremove, tContains, tonumber, tostring, strlen, type, date, time, select, unpack, error =
	  bit, math, table, pairs, ipairs, tinsert, tremove, tContains, tonumber, tostring, strlen, type, date, time, select, unpack, error
local UnitName, UnitGUID, GetPlayerInfoByGUID, GetUnitName, IsInGuild, SendAddonMessage =
	  UnitName, UnitGUID, GetPlayerInfoByGUID, GetUnitName, IsInGuild, SendAddonMessage
local WatchFrame, WatchFrame_OnEvent, ChatFrame1, ChatFrame_OnEvent, AlertFrame, AlertFrame_OnEvent =
	  WatchFrame, WatchFrame_OnEvent, ChatFrame1, ChatFrame_OnEvent, AlertFrame, AlertFrame_OnEvent
local ACHIEVEMENT_TOOLTIP_IN_PROGRESS, ACHIEVEMENT_TOOLTIP_COMPLETE, ACHIEVEMENT_BROADCAST =
	  ACHIEVEMENT_TOOLTIP_IN_PROGRESS, ACHIEVEMENT_TOOLTIP_COMPLETE, ACHIEVEMENT_BROADCAST
local ChatThrottleLib = ChatThrottleLib or {
	SendAddonMessage=function(_,_,prefix, message, channel, target)
		SendAddonMessage(prefix,message,channel,target)
	end,
}

local FIRST_CATEGORY = 1500
local CURRENT_CATEGORY

-- loaded flags for compability code
local pratLoaded

local sortFunc = function(a,b) return a.t > b.t end

local nameRealm = (UnitName("player")).." - "..GetRealmName()

local eventFrame = CreateFrame("Frame")

local registeredNamespaces = {}
local reverseBlizzAchievements = {}
local currentCompRecvData, currentCompToon = {}
local achievementDataCache = {}
local achievementStringCache = {}
local achievementRecv = {}
local compInfoCache = {}

local recentBlizzIndex = {}
local recentBlizzAchievements = {}
if UnitGUID("player") then
	recentBlizzAchievements.guid=UnitGUID("player")
	guidLoaded = true
else
	eventFrame:RegisterEvent("UNIT_NAME_UPDATE")
end
local scheduledBlizzMerge

local function mergeNSBlizz(namespaceTable)
	if (not namespaceTable.recentBlizz[nameRealm]) or (namespaceTable.recentBlizz[nameRealm].guid ~= UnitGUID("player")) then
		--print("GUID mismatch!",UnitGUID("player"),namespaceTable.recentBlizz[nameRealm].guid,nameRealm)
		namespaceTable.recentBlizz[nameRealm] = {guid=UnitGUID("player")}
	end
	--print("Merged:",namespaceTable,#namespaceTable.recentBlizz[nameRealm])
	for num, data in ipairs(namespaceTable.recentBlizz[nameRealm]) do
		if recentBlizzIndex[data.id] then
			recentBlizzIndex[data.id].t = math.min(recentBlizzIndex[data.id].t, data.t)
		else
			tinsert(recentBlizzAchievements,data)
			recentBlizzIndex[data.id] = data
		end
	end
	table.sort(recentBlizzAchievements,sortFunc)
	while #recentBlizzAchievements > 5 do
		recentBlizzIndex[tremove(recentBlizzAchievements).id] = nil
	end
	namespaceTable.recentBlizz[nameRealm] = recentBlizzAchievements
end

local function hookLatest(noLoadScheduled)
	local latestReturns, latestLoopIndexes = {}, {}
	local old_GetLatestCompletedAchievements = GetLatestCompletedAchievements
	function GetLatestCompletedAchievements(...)
		if (...) then
			return old_GetLatestCompletedAchievements(...)
		else
			table.wipe(latestLoopIndexes)
			table.wipe(latestReturns)
			for num=1,5 do
				local sourceNS = ""
				local latestTime = recentBlizzAchievements[latestLoopIndexes[""] or 1]
				for nsLabel, nsData in pairs(registeredNamespaces) do
					local comparedTable = nsData.saved.recent[latestLoopIndexes[nsLabel] or 1]
					if (comparedTable and comparedTable.t or 0) > (latestTime and latestTime.t or 0) then
						sourceNS = nsLabel
						latestTime = comparedTable
					end
				end
				if sourceNS == "" then
					latestReturns[num] = latestTime and latestTime.id
					latestLoopIndexes[""] = (latestLoopIndexes[""] or 1) + 1
				else
					latestReturns[num] = latestTime and (sourceNS.."_"..latestTime.id)
					latestLoopIndexes[sourceNS] = (latestLoopIndexes[sourceNS] or 1) + 1
				end
			end
			return unpack(latestReturns)
		end
	end
	if guidLoaded then
		isLoaded = true
		if not noLoadScheduled then
			for _,nsTab in ipairs(scheduledBlizzMerge) do
				--print("Merging:",nsTab,#nsTab.recentBlizz[nameRealm])
				mergeNSBlizz(nsTab)
			end
		end
	else
		recentLoaded = true
	end
end

local recentAchievements = {GetLatestCompletedAchievements()}
if #recentAchievements >= 1 then
	for avNum, avID in ipairs(recentAchievements) do
		recentBlizzAchievements[avNum] = {t=time()-avNum, id=avID}
		recentBlizzIndex[avID] = recentBlizzAchievements[avNum]
	end
	hookLatest(true)
else
	eventFrame:RegisterEvent("TRACKED_ACHIEVEMENT_UPDATE")
end
if not isLoaded then
	scheduledBlizzMerge = {}
end

local function ShowAchievementPopup(blizzID)
	AlertFrame_OnEvent(AlertFrame, "ACHIEVEMENT_EARNED", blizzID)
end

eventFrame:SetScript("OnEvent",function(self,event,...)
	if event == "CHAT_MSG_ADDON" then
		local prefix, message, chan, send = ...
		if prefix == "LCA ACH" and chan == "GUILD" then
			local nsLabel, achID = message:match("^(.-)_(%d+)")
			local ns = registeredNamespaces[nsLabel]
			local doRequest
			if not achievementDataCache[nsLabel.."_"..achID] then
				if not ns then
					doRequest = true
				else
					if not ns.reverseAchievements[tonumber(achID)] then
						doRequest=true
					end
				end
			end
			if doRequest then
				SendAddonMessage("LCA ADAT REQ",nsLabel.."_"..achID,"WHISPER",send)
			end
			local chatNum = 1
			local chatF = ChatFrame1
			while chatF do
				if pratLoaded then
					-- causes compare number with string error with regular code
					local message = ACHIEVEMENT_BROADCAST:format("","|cffffff00|Hachievement:"..message.."|h|r")
					while message:sub(1,1) == " " do message = message:sub(2) end
					ChatFrame_OnEvent(chatF, "CHAT_MSG_GUILD", message,send,GetDefaultLanguage(),"",send,"",0,0,"",0,0,"0x"..message:match("[^:]+:([^:]+)"),0)
				else
					ChatFrame_OnEvent(chatF, "CHAT_MSG_GUILD_ACHIEVEMENT", ACHIEVEMENT_BROADCAST:format("%s","|cffffff00|Hachievement:"..message.."|h|r"),send,send,"",0,0,"",0,0,"0x"..message:match("[^:]+:([^:]+)"),0)
				end
				chatNum = chatNum+1
				chatF = _G["ChatFrame"..chatNum]
			end
		end
		if prefix == "LCA ADAT REQ" and chan == "WHISPER" then
			local avData = achievementStringCache[message]
			if not avData then
				local nsLabel, achID = message:match("^(.-)_(%d+)")
				local ns = registeredNamespaces[nsLabel]
				if ns then
					local achievementData = ns.reverseAchievements[tonumber(achID)]
					avData = "|name|"..achievementData.name.."| |desc|"..achievementData.desc.."| |criteria|"
					for critNum,critFunc in ipairs(achievementData.criteria) do
						local label=critFunc(ns)
						avData = avData..critNum..":"..label..":"
					end
					avData = avData.."|"
					achievementStringCache[message] = avData
				end
			end
			local lengthEach = 250-strlen(message)
			local dataStrings = {}
			while strlen(avData) > lengthEach do
				tinsert(dataStrings,avData:sub(1,lengthEach))
				avData = avData:sub(lengthEach+1)
			end
			tinsert(dataStrings,avData)
			tinsert(dataStrings,"END")
			for strNum, strContent in ipairs(dataStrings) do
				ChatThrottleLib:SendAddonMessage("BULK","LCA ADAT DAT", message.." "..("%.03d"):format(strNum).." "..strContent, "WHISPER", send)
			end
		end
		if prefix == "LCA ADAT DAT" and chan == "WHISPER" then
			local id, num, content = message:match("^(.-_%d+) (%d+) (.+)")
			num = tonumber(num)
			achievementRecv[id] = achievementRecv[id] or {}
			local recvTable = achievementRecv[id]
			recvTable[num] = content
			if recvTable[#recvTable] == "END" then
				local completedData = table.concat(recvTable)
				local name, desc, critData = completedData:match("|name|([^|]+)| |desc|([^|]+)| |criteria|([^|]*)|")
				if (not name) or (not desc) then return end
				achievementDataCache[id] = {
					name=name,
					desc=desc,
					criteria = {
					},
				}
				for critNum, critName in critData:gmatch("(%d+):([^:]+):") do
					achievementDataCache[id].criteria[tonumber(critNum)] = critName
				end
			end
		end
		if prefix == "LCA COMP REQ" and chan == "WHISPER" then
			local avDataString = ""
			for nsLabel, nsData in pairs(registeredNamespaces) do
				local nsString
				for avID, avTime in pairs(nsData.saved.achievements) do
					nsString = (nsString and nsString.." " or "")..avID..":"..avTime
				end
				avDataString = avDataString..nsLabel.."$"..(nsString or "").."|"
			end
			local dataStrings = {}
			while strlen(avDataString) > 250 do
				tinsert(dataStrings,avDataString:sub(1,250))
				avDataString = avDataString:sub(251)
			end
			tinsert(dataStrings,avDataString)
			tinsert(dataStrings,"END")
			for strNum, strContent in ipairs(dataStrings) do
				ChatThrottleLib:SendAddonMessage("BULK","LCA COMP DAT", ("%.03d"):format(strNum).." "..strContent, "WHISPER", send)
			end
		end
		if prefix == "LCA COMP DAT" and chan == "WHISPER" then
			local entry, msg = message:match("^(%d+) (.+)")
			if entry then
				if currentCompToon == send then
					currentCompRecvData[tonumber(entry)] = msg
					if currentCompRecvData[#currentCompRecvData] == "END" then
						local completeString = table.concat(currentCompRecvData)
						if not compInfoCache[send] then compInfoCache[send] = {} end
						local compCache = compInfoCache[send]
						compCache.namespaces = compCache.namespaces and table.wipe(compCache.namespaces) or {}
						for nsLabel, nsData in completeString:gmatch("([^$]+)$([^|]*)|") do
							compCache.namespaces[nsLabel] = {}
							for avID, avTime in completeString:gmatch("(%d+):(%d+)") do
								compCache[nsLabel.."_"..avID] = tonumber(avTime)
								if registeredNamespaces[nsLabel] and registeredNamespaces[nsLabel].reverseAchievements and registeredNamespaces[nsLabel].reverseAchievements[tonumber(avID)] then
									compCache.namespaces[nsLabel][registeredNamespaces[nsLabel].reverseAchievements[tonumber(avID)].category] = (compCache.namespaces[nsLabel][registeredNamespaces[nsLabel].reverseAchievements[tonumber(avID)].category] or 0) + 1
								end
							end
						end
						AchievementFrameComparison_ForceUpdate()
					end
				end
			end
		end
	end
	if event == "ACHIEVEMENT_EARNED" then
		-- GLOBALS: GetAchievementInfo
		if (select(12,GetAchievementInfo((...)))) then return end
		local id, t = (...), time()
		if recentBlizzIndex[id] then
			recentBlizzIndex[id].t = t
			table.sort(recentBlizzAchievements,sortFunc)
		else
			tinsert(recentBlizzAchievements,1,{id=id,t=t})
			local removedAchiev = tremove(recentBlizzAchievements,6)
			if removedAchiev then
				recentBlizzIndex[removedAchiev.id] = nil
			end
		end
	end
	if event == "TRACKED_ACHIEVEMENT_UPDATE" then
		self:UnregisterEvent("TRACKED_ACHIEVEMENT_UPDATE")
		for avNum, avID in ipairs({GetLatestCompletedAchievements()}) do
			recentBlizzAchievements[avNum] = {t=time()-avNum, id=avID}
			recentBlizzIndex[avID] = recentBlizzAchievements[avNum]
		end
		hookLatest()
	end
	if event == "UNIT_NAME_UPDATE" then
		if UnitIsUnit((...),"player") then
			recentBlizzAchievements.guid = UnitGUID("player")
			self:UnregisterEvent("UNIT_NAME_UPDATE")
			if recentLoaded then
				isLoaded = true
				for _,nsTab in ipairs(scheduledBlizzMerge) do
					--print("Merging:",nsTab,#nsTab.recentBlizz[nameRealm])
					mergeNSBlizz(nsTab)
				end
			else
				guidLoaded = true
			end
		end
	end
	if event == "ADDON_LOADED" then
		if (...) == "Prat-3.0" then
			pratLoaded = true
		end
	end
end)
eventFrame:RegisterEvent("CHAT_MSG_ADDON")
RegisterAddonMessagePrefix("LCA ACH")
RegisterAddonMessagePrefix("LCA ADAT REQ")
RegisterAddonMessagePrefix("LCA ADAT DAT")
RegisterAddonMessagePrefix("LCA COMP REQ")
RegisterAddonMessagePrefix("LCA COMP DAT")
eventFrame:RegisterEvent("ACHIEVEMENT_EARNED")
eventFrame:RegisterEvent("CRITERIA_UPDATE")
eventFrame:RegisterEvent("ADDON_LOADED")
--eventFrame:RegisterEvent("TRACKED_ACHIEVEMENT_UPDATE")

local namespaceFuncs = {
	RegisterAchievement = function(ns, id, name, desc, icon, criteria, category, prevAch, nextAch)
		if not ns.achievementGroups[category] then
			ns.achievementGroups[category] = {}
			if not (category == "") then
				CURRENT_CATEGORY = CURRENT_CATEGORY or (FIRST_CATEGORY+1)
				reverseBlizzAchievements[CURRENT_CATEGORY] = {category,ns.rootCategoryID,0,ns=ns}
				ns.blizzAchievements[category] = CURRENT_CATEGORY
				CURRENT_CATEGORY = CURRENT_CATEGORY+1
			end
		end
		if prevAch then
			local prevData = ns.reverseAchievements[prevAch]
			tinsert(ns.achievementGroups[prevData.category][prevData.idInCat], {id=id,icon=icon,criteria=criteria,desc=desc,name=name,prevAch=prevAch,nextAch=nextAch})
			ns.reverseAchievements[id] = {category=prevData.category, idInCat=prevData.idInCat, icon=icon, criteria=criteria, desc=desc, name=name, prevAch=prevAch, nextAch=nextAch}
		else
			tinsert(ns.achievementGroups[category], {{id=id, icon=icon, criteria=criteria, desc=desc, name=name, nextAch=nextAch}})
			ns.reverseAchievements[id] = {category=category, idInCat=#ns.achievementGroups[category], icon=icon, criteria=criteria, desc=desc, name=name, nextAch=nextAch}
		end
	end,
	RegenerateAchievementCategoryListings=function(ns)
		--[[if not eventFrame.isLoaded then
			eventFrame.scheduledRegenerate[ns.label] = true
			return
		end--]]
		for category, data in pairs(ns.achievementGroups) do
			ns.categoryListingCache[category] = ns.categoryListingCache[category] and table.wipe(ns.categoryListingCache[category]) or {}
			local cache = ns.categoryListingCache[category]
			local completed = 0
			for _, progTable in ipairs(data) do
				for i=#progTable, 1, -1 do
					local avData = progTable[i]
					if ns:HasAchievement(avData.id) then
						tinsert(cache, avData.id)
						completed = completed+1
						break
					end
				end
			end
			for _, progTable in ipairs(data) do
				for _, avData in ipairs(progTable) do
					if (not ns:HasAchievement(avData.id)) then
						tinsert(cache, avData.id)
						break
					end
				end
			end
			cache.numCompleted = completed
		end
		if not ns.categoryListingCache[""] then
			ns.categoryListingCache[""] = {numCompleted = 0}
		end
		if AchievementFrame.selectedTab == 1 then
			AchievementFrameCategories_GetCategoryList(ACHIEVEMENTUI_CATEGORIES)
		end
		AchievementFrameCategories_Update()
	end,
	HasAchievement=function(ns,id)
		if ns.saved.achievements and ns.saved.achievements[id] then
			return true
		end
	end,
	AwardAchievement=function(ns,id)
		local blizzID = ns.label.."_"..id
		ns.saved.achievements = ns.saved.achievements or {}
		ns.saved.achievements[id] = time()
		tinsert(ns.saved.recent,1,{t=time(),id=id})
		tremove(ns.saved.recent,6)
		ns.saved.tracked[id] = nil
		ShowAchievementPopup(blizzID)
		ns:RegenerateAchievementCategoryListings()
		AchievementFrame_ForceUpdate()
		AchievementFrameAchievements_OnEvent(AchievementFrameAchievements, "TRACKED_ACHIEVEMENT_UPDATE")
		WatchFrame_OnEvent(WatchFrame, "TRACKED_ACHIEVEMENT_UPDATE")
		if IsInGuild() then
			-- GLOBALS: GetAchievementLink
			SendAddonMessage("LCA ACH", GetAchievementLink(blizzID):match("|cffffff00|Hachievement:(.+)|h|r"),"GUILD")
			if ns.saved.announceAchievement then
				local data = ns.reverseAchievements[id]
				if data then
					SendChatMessage("[LCA] "..(UnitName("player")).." has earned the achievement ["..data.name.."] (\""..data.desc.."\")", "GUILD")
				end
			end
		end
	end,
	WriteAchievementMemory=function(ns,avID,index,value)
		local nsTable = ns.saved
		if value then
			nsTable.memory[avID] = bit.bor(nsTable.memory[avID] or 0, (2^(index-1)))
		else
			local bString = nsTable[avID] or 0
			nsTable.memory[avID] = bString-bit.band(bString,(2^(index-1)))
		end
	end,
	ReadAchievementMemory=function(ns,avID,index)
		return (bit.band(ns.saved.memory[avID] or 0,(2^(index-1)))>0)
	end,
	SetCategoryOrder=function(ns,order)
		ns.categoryOrder = order
	end,
}
----------------------------------
-- :RegisterNamespace arguments --
----------------------------------
-- namespaceLabel [string]
---- unique label for the namespace - used in the achievement window as well as in achievement IDs
-- namespaceTable [table]
---- a table containing achievement memory and achievement history for the namespace. storing this cross-session is the embedding addon's responsibility.
----- the type of savedVar used also determines whether achievements are per character or per realm
function lib:RegisterNamespace(namespaceLabel, namespaceTable)
	if namespaceLabel:find("_%d+") then
		error("Invalid namespace label (contains <_num>)")
	end
	if registeredNamespaces[namespaceLabel] then
		error("Namespace already exists: "..tostring(namespaceLabel))
	end
	local namespace = {}
	CURRENT_CATEGORY = CURRENT_CATEGORY or (FIRST_CATEGORY+1)
	namespace.rootCategoryID = CURRENT_CATEGORY
	namespace.reverseAchievements = {}
	namespace.blizzAchievements = {[""] = CURRENT_CATEGORY}
	reverseBlizzAchievements[CURRENT_CATEGORY] = {namespaceLabel,-1,0,ns=namespace}
	namespace.achievementGroups = {}
	namespace.categoryListingCache = {}
	CURRENT_CATEGORY = CURRENT_CATEGORY+1
	namespace.label = namespaceLabel
	namespace.saved = namespaceTable
	if not namespaceTable.wasLoadedAlready then
		-- init NS table
		namespaceTable.achievements = {}
		namespaceTable.recent = {}
		namespaceTable.tracked = {}
		namespaceTable.memory = {}
		namespaceTable.recentBlizz = {}
		namespaceTable.wasLoadedAlready = true
	end
	-- process NS table
	if isLoaded then
		mergeNSBlizz(namespaceTable)
	else
		--print("Scheduling:",namespaceTable,#namespaceTable.recentBlizz[nameRealm])
		tinsert(scheduledBlizzMerge,namespaceTable)
	end
	-- import funcs
	for key,func in pairs(namespaceFuncs) do
		namespace[key] = func
	end
	registeredNamespaces[namespaceLabel] = namespace
	return namespace
end

ChatFrame_AddMessageEventFilter("CHAT_MSG_GUILD", function(self,event,msg)
	if msg:match("^%[LCA%]") then
		return true
	end
end)
LoadAddOn("Blizzard_AchievementUI")
hooksecurefunc("AchievementFrameCategories_GetCategoryList", function(t)
	-- while comparing, selectedTab is set after we're called - if tab 1 is selected, selectedTab will thus be 3 (tab 2 is hidden during comparison)
	-- also, selectedTab being nil means that the frame is in the process of being opened, which will open category 1
	local isComp = AchievementFrame_IsComparison()
	if (not AchievementFrame.selectedTab) or ((not isComp) and (AchievementFrame.selectedTab == 1)) or (isComp and (AchievementFrame.selectedTab == 3)) then
		for nsLabel, ns in pairs(registeredNamespaces) do
			tinsert(t, (#t), {id=ns.rootCategoryID})
			if ns.categoryOrder then
				for _, category in ipairs(ns.categoryOrder) do
					if ns.blizzAchievements[category] then
						tinsert(t, (#t), {id=ns.blizzAchievements[category], parent=ns.rootCategoryID, hidden=true})
					end
				end
				for name, blizzID in pairs(ns.blizzAchievements) do
					if (name ~= "") and (not tContains(ns.categoryOrder,name)) then
						tinsert(t, (#t), {id=blizzID, parent=ns.rootCategoryID, hidden=true})
					end
				end
			else
				for name, blizzID in pairs(ns.blizzAchievements) do
					if name ~= "" then
						tinsert(t, (#t), {id=blizzID, parent=ns.rootCategoryID, hidden=true})
					end
				end
			end
		end
	end
end)
hooksecurefunc("SetAchievementComparisonUnit", function(uID)
	local name = GetUnitName(uID):gsub(" ","")
	table.wipe(currentCompRecvData)
	currentCompToon = name
	SendAddonMessage("LCA COMP REQ","","WHISPER",name)
end)

--hook various functions
--[[local old_tonumber = tonumber
-- disabled, taints in various places
function tonumber(arg1,...)
	if type(arg1) == "string" then
		local nsName = arg1:match("(.+)_%d+")
		if registeredNamespaces[nsName] then
			return arg1,old_tonumber(...)
		end
	end
	return old_tonumber(arg1,...)
end--]]

-- remove custom achievement links from SendChatMessage - server would otherwise suppress the message
-- BN whispers allow them, but they also use BNSendWhisper
-- GLOBALS: SendChatMessage
local old_SendChatMessage = SendChatMessage
function SendChatMessage(msg,...)
	for nsName in pairs(registeredNamespaces) do
		msg = msg:gsub("|cffffff00|Hachievement:"..nsName..".-|h(%[.-%])|h|r","%1")
	end
	old_SendChatMessage(msg,...)
end

local old_GetCategoryInfo = GetCategoryInfo
function GetCategoryInfo(id)
	if reverseBlizzAchievements[id] then
		return unpack(reverseBlizzAchievements[id])
	else
		if not id then
			return false
		end
		return old_GetCategoryInfo(id)
	end
end

local old_GetCategoryNumAchievements = GetCategoryNumAchievements
function GetCategoryNumAchievements(...)
	local revTable = reverseBlizzAchievements[(...)]
	if revTable then
		local ns = revTable.ns
		local catName = (ns.rootCategoryID == (...)) and "" or revTable[1]
		local showAll = select(2,...)
		if showAll then
			if not ns.achievementGroups[catName] then return 0,0 end
			local completed, all = 0,0
			for _, d in pairs(ns.achievementGroups[catName]) do
				all = all+#d
				for _, a in ipairs(d) do
					if ns:HasAchievement(a.id) then
						completed = completed+1
					end
				end
			end
			return all, completed
		else
			if not ns.categoryListingCache[catName] then
				ns:RegenerateAchievementCategoryListings()
			end
			return #ns.categoryListingCache[catName], ns.categoryListingCache[catName].numCompleted
		end
	else
		return old_GetCategoryNumAchievements(...)
	end
end

local old_GetAchievementInfo = GetAchievementInfo
function GetAchievementInfo(...)
	local achievID
	local achievNS
	if select('#',...) == 2 then
		local id = select(1,...)
		local revTable = reverseBlizzAchievements[id]
		if revTable then
			local ns = revTable.ns
			local catName = (ns.rootCategoryID == id) and "" or revTable[1]
			if catName then
				if not ns.categoryListingCache[catName] then
					ns:RegenerateAchievementCategoryListings()
				end
				achievID = ns.categoryListingCache[catName][(select(2,...))]
				achievNS = ns
				if not achievID then
					return nil
				end
			end
		end
	else
		local id = (...)
		if type(id) == "string" then
			local nsLabel, ownID = id:match("(.+)_(%d+)")
			if registeredNamespaces[nsLabel] then
				achievID = tonumber(ownID)
				achievNS = registeredNamespaces[nsLabel]
			end
		end
	end
	if achievID then
		local achievData = achievNS.reverseAchievements[achievID]
		if not achievData then
			return -1, "Invalid achievement", 0, false, nil, nil, nil, "Invalid achievement", 0, "Interface\\Icons\\Inv_Misc_QuestionMark", "", false
		end
		local completionDate = achievNS.saved.achievements[achievID]
		if completionDate then
			local dateInfo = date("*t", completionDate)
			return achievNS.label.."_"..achievID, achievData.name, 0, true, dateInfo.month, dateInfo.day, (dateInfo.year%100), achievData.desc, 0, achievData.icon, "", false
		else
			return achievNS.label.."_"..achievID, achievData.name, 0, false, nil, nil, nil, achievData.desc, 0, achievData.icon, "", false
		end
	else
		return old_GetAchievementInfo(...)
	end
end

local old_IsTrackedAchievement = IsTrackedAchievement
function IsTrackedAchievement(id)
	if type(id) == "string" then
		local nsLabel, achievID = id:match("(.+)_(%d+)")
		if nsLabel and registeredNamespaces[nsLabel] then
			return registeredNamespaces[nsLabel].saved.tracked[tonumber(achievID)]
		end
	end
	return old_IsTrackedAchievement(id)
end

local old_AddTrackedAchievement = AddTrackedAchievement
function AddTrackedAchievement(id)
	if type(id) == "string" then
		local nsLabel, achievID = id:match("(.+)_(%d+)")
		if nsLabel and registeredNamespaces[nsLabel] then
			registeredNamespaces[nsLabel].saved.tracked[tonumber(achievID)] = true
			AchievementFrameAchievements_OnEvent(AchievementFrameAchievements, "TRACKED_ACHIEVEMENT_UPDATE")
			WatchFrame_OnEvent(WatchFrame, "TRACKED_ACHIEVEMENT_UPDATE")
			return true
		end
	end
	return old_AddTrackedAchievement(id)
end

local old_RemoveTrackedAchievement = RemoveTrackedAchievement
function RemoveTrackedAchievement(id)
	if type(id) == "string" then
		local nsLabel, achievID = id:match("(.+)_(%d+)")
		if nsLabel and registeredNamespaces[nsLabel] then
			registeredNamespaces[nsLabel].saved.tracked[tonumber(achievID)] = nil
			AchievementFrameAchievements_OnEvent(AchievementFrameAchievements, "TRACKED_ACHIEVEMENT_UPDATE")
			WatchFrame_OnEvent(WatchFrame, "TRACKED_ACHIEVEMENT_UPDATE")
			return true
		end
	end
	return old_RemoveTrackedAchievement(id)
end

local old_GetAchievementNumCriteria = GetAchievementNumCriteria
function GetAchievementNumCriteria(id)
	if type(id) == "string" then
		local nsLabel, achievID = id:match("(.+)_(%d+)")
		if nsLabel and registeredNamespaces[nsLabel] then
			return (#registeredNamespaces[nsLabel].reverseAchievements[tonumber(achievID)].criteria)
		end
	end
	return old_GetAchievementNumCriteria(id)
end

local old_GetAchievementCriteriaInfo = GetAchievementCriteriaInfo
function GetAchievementCriteriaInfo(...)
	if select('#', ...) == 2 then
		local avID, index = select(1,...)
		if type(avID) == "string" then
			local nsLabel, achievID = avID:match("(.+)_(%d+)")
			if nsLabel and registeredNamespaces[nsLabel] then
				local ns = registeredNamespaces[nsLabel]
				local arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9 = ns.reverseAchievements[tonumber(achievID)].criteria[index](ns)
				return arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, avID.."_"..index
			end
		end
	end
	return old_GetAchievementCriteriaInfo(...)
end

local old_GetAchievementCategory = GetAchievementCategory
function GetAchievementCategory(id)
	if type(id) == "string" then
		local nsLabel, achievID = id:match("(.+)_(%d+)")
		if nsLabel and registeredNamespaces[nsLabel] then
			local ns = registeredNamespaces[nsLabel]
			return ns.blizzAchievements[ns.reverseAchievements[tonumber(achievID)].category]
		end
	end
	if not id then
		return false
	end
	return old_GetAchievementCategory(id)
end

local old_GetPreviousAchievement = GetPreviousAchievement
function GetPreviousAchievement(id)
	if type(id) == "string" then
		local nsLabel, achievID = id:match("(.+)_(%d+)")
		if nsLabel and registeredNamespaces[nsLabel] then
			local prevAch = registeredNamespaces[nsLabel].reverseAchievements[tonumber(achievID)].prevAch
			if prevAch then
				return nsLabel.."_"..prevAch
			else
				return
			end
		end
	end
	return old_GetPreviousAchievement(id)
end

local old_GetNextAchievement = GetNextAchievement
function GetNextAchievement(id)
	if type(id) == "string" then
		local nsLabel, achievID = id:match("(.+)_(%d+)")
		if nsLabel and registeredNamespaces[nsLabel] then
			local nextAch = registeredNamespaces[nsLabel].reverseAchievements[tonumber(achievID)].nextAch
			if nextAch then
				return nsLabel.."_"..nextAch
			else
				return
			end
		end
	end
	return old_GetNextAchievement(id)
end

local old_GetAchievementLink = GetAchievementLink
function GetAchievementLink(id)
	if type(id) == "string" then
		local nsLabel, achievID = id:match("(.+)_(%d+)")
		if nsLabel and registeredNamespaces[nsLabel] then
			achievID = tonumber(achievID)
			local ns = registeredNamespaces[nsLabel]
			local completed = ns.saved.achievements[achievID]
			local flag = 0
			for id, func in ipairs(ns.reverseAchievements[achievID]) do
				if (select(3, func(ns))) then
					flag=flag+(2^(id-1))
				end
			end
			if completed then
				local d = date("*t", completed)
				return "|cffffff00|Hachievement:"..id..":"..UnitGUID("player"):sub(3)..":1:"..d.month..":"..d.day..":"..(d.year%100)..":"..flag..":0:0|h["..ns.reverseAchievements[achievID].name.."]|h|r"
			else
				return "|cffffff00|Hachievement:"..id..":"..UnitGUID("player"):sub(3)..":0:0:0:-1:"..flag..":0:0|h["..ns.reverseAchievements[achievID].name.."]|h|r"
			end
		end
	end
	return old_GetAchievementLink(id)
end

local old_GetTrackedAchievements = GetTrackedAchievements
function GetTrackedAchievements(...)
	local trackedAchievements = {old_GetTrackedAchievements(...)}
	for nsLabel, ns in pairs(registeredNamespaces) do
		for avID, doTrack in pairs(ns.saved.tracked) do
			if doTrack then
				tinsert(trackedAchievements,nsLabel.."_"..avID)
			end
		end
	end
	return unpack(trackedAchievements)
end

local old_GetNumTrackedAchievements = GetNumTrackedAchievements
function GetNumTrackedAchievements(...)
	local i=old_GetNumTrackedAchievements(...)
	for nsLabel, ns in pairs(registeredNamespaces) do
		for avID, doTrack in pairs(ns.saved.tracked) do
			if doTrack then
				i=i+1
			end
		end
	end
	return i
end

local old_IsAchievementEligible = IsAchievementEligible
function IsAchievementEligible(id)
	if type(id) == "string" then
		local nsLabel, achievID = id:match("(.+)_(%d+)")
		if nsLabel and registeredNamespaces[nsLabel] then
			return
		end
	end
	return old_IsAchievementEligible(id)
end

local old_GetAchievementComparisonInfo = GetAchievementComparisonInfo
function GetAchievementComparisonInfo(...)
	if type((...)) == "string" then
		local compCache = compInfoCache[currentCompToon]
		local nsLabel, achievID = (...):match("^(.-)_(%d+)")
		if (compCache and compCache.namespaces[nsLabel]) or registeredNamespaces[nsLabel] then
			if compCache and compCache[(...)] then
				local dateInfo = date("*t", compCache[(...)])
				return true, dateInfo.month, dateInfo.day, (dateInfo.year%100)
			else
				return nil, nil, nil, nil
			end
		end
	end
	return old_GetAchievementComparisonInfo(...)
end

local old_GetComparisonCategoryNumAchievements = GetComparisonCategoryNumAchievements
function GetComparisonCategoryNumAchievements(...)
	local id = ...
	if reverseBlizzAchievements[id] then
		local ns = reverseBlizzAchievements[id].ns
		if compInfoCache[currentCompToon] and compInfoCache[currentCompToon].namespaces and compInfoCache[currentCompToon].namespaces[ns.label] then
			return compInfoCache[currentCompToon].namespaces[ns.label][reverseBlizzAchievements[id][1]] or 0
		else
			return 0
		end
	end
	return old_GetComparisonCategoryNumAchievements(...)
end

--ASDFWASD = compInfoCache

local old_GetStatistic = GetStatistic
function GetStatistic(...)
	if type((...)) == "string" then
		local nsLabel, achievID = (...):match("(.+)_(%d+)")
		if nsLabel and registeredNamespaces[nsLabel] then
			return "0"
		end
	end
	return old_GetStatistic(...)
end

hooksecurefunc(ItemRefTooltip, "SetHyperlink", function(self, link)
	local ownID, unitGUID, completed, month, day, year, flags = link:match("achievement:(.-_%d+):([0-9A-Fa-f]+):([0-1]):(%d+):(%d+):([%-0-9]+):(%d+):0:0")
	if ownID then
		local complete = (completed == "1")
		local nsLabel, achievID = ownID:match("^(.-)_(%d+)")
		local ns = registeredNamespaces[nsLabel]
		local achievData = ns and ns.reverseAchievements[tonumber(achievID)]
		local isCached = false
		if not achievData then
			achievData = achievementDataCache[ownID]
			isCached = true
		end
		if achievData then
			self:ClearLines()
			self:AddLine(achievData.name,1,1,1,false)
			self:AddLine(" ")
			local playerName = (select(6,GetPlayerInfoByGUID(unitGUID))) or "Unknown"
			if complete then
				self:AddLine(ACHIEVEMENT_TOOLTIP_COMPLETE:format(playerName, day, month, year),0,1,0)
			else
				self:AddLine(ACHIEVEMENT_TOOLTIP_IN_PROGRESS:format(playerName),0,1,0)
			end
			self:AddLine(" ")
			self:AddLine(achievData.desc,1,1,1,true)
			self:AddLine(" ")
			flags = tonumber(flags)
			for critCnt=1, #achievData.criteria, 2 do
				local label1
				if isCached then
					label1 = achievData.criteria[critCnt]
				else
					label1 = achievData.criteria[critCnt](ns)
				end
				local has1 = (bit.band(flags,(2^(critCnt-1)))>0)
				local label2, has2
				if #achievData.criteria >= (critCnt+1) then
					if isCached then
						label2 = achievData.criteria[critCnt+1]
					else
						label2 = achievData.criteria[critCnt+1](ns)
					end
					has2 = (bit.band(flags,(2^critCnt))>0)
				else
					label2 = " "
					has2 = false
				end
				self:AddDoubleLine(label1, label2, has1 and 0.25 or 0.5, has1 and 0.75 or 0.5, has1 and 0.25 or 0.5, has2 and 0.25 or 0.5, has2 and 0.75 or 0.5, has2 and 0.25 or 0.5)
			end
			self:Show()
		end
	end
end)